--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2023 Adobe
All Rights Reserved.

NOTICE: All information contained herein is, and remains
the property of Adobe and its suppliers, if any. The intellectual
and technical concepts contained herein are proprietary to Adobe
and its suppliers and are protected by all applicable intellectual
property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material
is strictly forbidden unless prior written permission is obtained
from Adobe.

------------------------------------------------------------------------------]]

	-- Rx
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.utils
local KSData = require 'KSData'
local KSIdGenerator = require 'KSIdGenerator'
local KSPathUtils = require 'KSPathUtils'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFImsAppIdentity = require 'WFImsAppIdentity'
local WFFileUtils = require 'WFFileUtils'
local WFLogger = require 'WFLogger'

	-- WF.web
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'
local WFRxHttpClient = require 'WFRxHttpClient'

--------------------------------------------------------------------------------

local logger = WFLogger( 'WFCatOrbit' )
local check = assert( WFArguments.check )

local KSTableUtils_derefOpt = KSTableUtils.derefOpt

--------------------------------------------------------------------------------

-- Sensei

local kSenseiV2Predict_Stage = "https://sensei-stage-ue1.adobe.io/services/v2/predict"
local kSenseiV2Predict_Prod = "https://sensei-ue1.adobe.io/services/v2/predict"

-- Autotag Service Constants

local kAutoTagServiceUrl_Prod = kSenseiV2Predict_Prod
local kAutotagServiceName = 'Classification:autotags:Service-011d3ac6049246c9b54d2bcd456c1264'
local kAutotagServiceTagger = 'tagger'
local kAutotagServiceCategory = 'category'
local kAutotagMaxResultsDefault = 100

-- Firefly Service Constants

local kFireflyServiceUrl_Prod = "https://firefly.adobe.io/spl"
local kFireflyServiceUrl_Stage = "https://firefly-stage.adobe.io/spl"
local kOutpaintingServiceGraph = 'urn:graph:MultiDiffusion_outpaint_v3'
local kOutpaintingModeGoup = 'goup' -- generic outpainting
local kOutpaintingModeToup = 'toup' -- text-guided outpainting

-- MetaCAF Service Constants

local kDIIMetaCAFUrl_Stage = kSenseiV2Predict_Stage
local kDIIMetaCAFUrl_Prod = kSenseiV2Predict_Prod
local kDIIMetaCAFServiceName = "Feature:autocrop:Service-f0372bb81536409598142277cbb2bf34"
local kDIIMetaCAFServiceId = "Service-f0372bb81536409598142277cbb2bf34"

-- DII SSM CPF Service Constants

local kDIIServerSideMaskingUrl_Stage = kSenseiV2Predict_Stage
local kDIIServerSideMaskingUrl_Prod = kSenseiV2Predict_Prod
local kDIIServerSideMaskingServiceName = "Classification:autocrop:Service-e02a518da69e44c7a90ea5f1006a02bd"

--============================================================================--

local WFCatOrbit = {}

--------------------------------------------------------------------------------
--@mark Sensei Prediction Services
--------------------------------------------------------------------------------

--[[
	The following code was written to explore Sensei integrations within WF.
	It is unused, and outside of the unit tests, it is untested. It can be
	used as a reference, or for further experimentation.
]]--

local function requestAutotagsForImage( authToken, filePath, autotagService, maxResults )

	-- https://wiki.corp.adobe.com/pages/viewpage.action?pageId=2269164104

	logger:infof( "requestAutotagsForImage filePath: %s, autotagService %s", filePath, autotagService )
 
	autotagService = autotagService or kAutotagServiceTagger

	maxResults = maxResults or kAutotagMaxResultsDefault

	if not filePath or not WFFileUtils.exists( filePath ) then

		return Observable.throw(
				string.format( "The specified file does not exist: %s", filePath ) )

	end

	local imsClientId = WFImsAppIdentity.getClientId()

	local requestHeaders = {
		{
			header = 'X-Api-Key',
			value = imsClientId
		},
		{
			header = 'Prefer',
			value = 'respond-sync, wait=1'
		},
		{
			header = 'Cache-Control',
			value = 'no-cache'
		},
		{
			header = 'Authorization',
			value = string.format( "Bearer %s", authToken )
		}
	}

	local requestOptions = {
		requestMethod = 'POST',
		returnHttpErrorsAsData = true,
		requestHeaders = requestHeaders,
		withoutAdobeAuthorization = true,
		multipartRequest = true,
		requestCost = WFRxHttpClient.requestCostUploadNegative,
		requestPriority = WFRxHttpClient.requestPriorityHighest
	}

	local contentAnalyzerRequests = {
		[ "sensei:invocation_mode" ] = "synchronous",
		[ "sensei:invocation_batch" ] = false,
		[ "sensei:name" ] = kAutotagServiceName,
		[ "sensei:engines" ] = {
			{
				[ "sensei:execution_info" ] = {
					[ "sensei:engine" ] = kAutotagServiceName
				},
				[ "sensei:params" ] = {
					[ "num_search_hits" ] = maxResults,
					[ "service" ] = autotagService
				},
				[ "sensei:inputs" ] = {
					[ "image_input"] = {
						[ "dc:format" ] = "image/jpg",
						[ "sensei:multipart_field_name" ] = "infile"
					}
				},
				[ "sensei:outputs" ] = {
					[ "result" ] = {
						[ "dc:format" ] = "application/json",
						[ "sensei:multipart_field_name" ] = "classifications"
					}
				}
			}
		}
	}

	local fileName = KSPathUtils.leafName( filePath )

	local multipartRequestData = {
		{
			name = 'contentAnalyzerRequests',
			contentType = 'application/json',
			value = KSJsonEncoder.prettyPrint( contentAnalyzerRequests, { pretty = true } )
		},
		{
			name = 'infile',
			filePath = filePath,
			fileName = fileName
		}
	}

	return WFRxHttpClient.put( kAutoTagServiceUrl_Prod, multipartRequestData, requestOptions )

end

--------------------------------------------------------------------------------

--[[!

	\brief Fetches autotags from Sensei for a specified image.

	\param authToken (string) The authorization token needed to communicate with
		the sensei service.
	\param filePath (string) A file path to the image.
	\param maxResults (?number) The number of autotags the service should generate
		and return. Defaults to 100.

	\return Observable which fires with the HTTP response from the service.
		See https://wiki.corp.adobe.com/pages/viewpage.action?pageId=2269164104 for
		more information on the response.

--]]

function WFCatOrbit.requestAutotagsForImage_experimental( authToken, filePath, maxResults )

	check( "+string", "+string", '?number' )

	return requestAutotagsForImage( authToken, filePath, kAutotagServiceTagger, maxResults )

end

--------------------------------------------------------------------------------

--[[!
	\brief Fetches categorization autotags from Sensei for the specified file.

	\param authToken (string) The authorization token needed to communicate with
		the sensei service.
	\param filePath (string) A file path to the image.
	\param maxResults (?number) The number of autotags the service should generate
		and return. Defaults to 100.

	\return Observable which fires with the HTTP response from the service.
--]]

function WFCatOrbit.requestCategoryAutotagsForImage_experimental( authToken, filePath, maxResults )

	check( "+string", "+string", '?number' )

	return requestAutotagsForImage( authToken, filePath, kAutotagServiceCategory, maxResults )

end

--------------------------------------------------------------------------------
--@mark Firefly Service
--------------------------------------------------------------------------------

local readyForRandomness = false

local function prepRandomness()

	if not readyForRandomness then
	
		-- Produces a different sequence each time (assuming enough time between invocations)
	
		math.randomseed( os.time() )
		math.random()
		math.random()
		math.random()
		
		readyForRandomness = true
	
	end

end

--------------------------------------------------------------------------------

local function getSeedValue()

	-- Generate a random seed value. Max and Min values are defined by the service.

	local min = 0
	local max = 2147483647

	prepRandomness()

	local random = math.random( min, max - 1 )

	return random

end

--------------------------------------------------------------------------------

local function getSeedArray( count )

	local seeds = {}

		for i = 0, count - 1, 1 do

		local seedInfo = {
			name = tostring( i ),
			type = 'scalar',
			value = getSeedValue()
		}

		seeds[ #seeds + 1 ] = seedInfo

	end
	
	return seeds

end

--------------------------------------------------------------------------------

local sessionId = KSIdGenerator.uuidInLowercaseHexWithHyphens()
	-- https://lightroom.slack.com/archives/C04DUFT283H/p1684816671093209?thread_ts=1684815815.573139&cid=C04DUFT283H
	-- Clients must generate a random guid when a firefly session starts and use
	-- that same guid for as long as the session is active in the app.
	-- It is highly discouraged to create a new session id for every request as
	-- gaining a new session may be restricted if the service is above capacity.
	-- So reusing a session for as long as possible is best practice.

local function fireflyOutpaintingRequest( wfCatalog, filePath, size, optArgs )

	-- See https://wiki.corp.adobe.com/display/DISysEng/MultiDiffusion+Service#MultiDiffusionService-Outpainting(withAspectRatio)
	-- for more details on outpainting parameters, and requests/responses.

	logger:infof( "fireflyOutpaintingRequest filePath: %s, size %s, optArgs %s", filePath, size, optArgs )

	if not filePath or not WFFileUtils.exists( filePath ) then

		return Observable.throw(
				string.format( "The specified file does not exist: %s", filePath ) )

	end

	-- Prompt Processing

	local prompt = optArgs.prompt

	local mode = kOutpaintingModeGoup

	local maybePromptParam

	if prompt then
	
		mode = kOutpaintingModeToup
		
		maybePromptParam = {
			[ 'name' ] = "gi_PROMPT",
			[ 'value' ] = prompt,
			[ 'type' ] = "string"
		}
	
	end

	-- Form Firefly Request

	local request = {
	
		[ 'graph' ] = {
			[ 'uri' ] = kOutpaintingServiceGraph
		},
		
		[ 'params' ] = {
			{
				[ 'name' ] = "gi_NUM_STEPS",
				[ 'value' ] = 40, -- Number of sampling steps
				[ 'type' ] = "scalar"
			},
			{
				[ 'name' ] = "gi_MODE",
				[ 'value' ] = mode,
				[ 'type' ] = "string"
			},
			{
				[ 'name' ] = "gi_AR_SHIFT",
				[ 'value' ] = optArgs.shift or 0, -- Shift position for outpainting
				[ 'type' ] = "scalar"
			},
			{
				[ 'name' ] = "gi_AR_DILATE",
				[ 'value' ] = optArgs.dilate or 0, -- Dilate mask to have better boundary
				[ 'type' ] = "scalar"
			},
			{
				[ 'name' ] = "gi_SEED",
				[ 'value' ] = getSeedArray( 4 ),
				[ 'type' ] = "array"
			},
			{
				[ 'name' ] = "gi_OUTPUT_HEIGHT",
				[ 'value' ] = size.height,
				[ 'type' ] = "scalar"
			},
			{
				[ 'name' ] = "gi_OUTPUT_WIDTH",
				[ 'value' ] = size.width,
				[ 'type' ] = "scalar"
			},
			{
				[ 'name' ] = "gi_AR_CUSTOM_H",
				[ 'value' ] = size.height,
				[ 'type' ] = "scalar"
			},
			{
				[ 'name' ] = "gi_AR_CUSTOM_W",
				[ 'value' ] = size.width,
				[ 'type' ] = "scalar"
			},
			maybePromptParam
		},

		[ 'inputs' ] = {
			[ 'gi_IMAGE' ] = {
				[ 'id' ] = KSIdGenerator.uuidInLowercaseHexWithHyphens(),
				[ 'type' ] = "image",
			}
		},

		[ 'outputs' ] = {
			[ 'gi_GEN_IMAGE' ] = {
				[ 'id' ] = KSIdGenerator.uuidInLowercaseHexWithHyphens(),
				[ 'type' ] = "array",
				[ 'expectedMimeType' ] = "image/jpeg",
				[ 'expectedArrayLength' ] = 4,
			},
			[ 'gi_GEN_STATUS' ] = {
				[ 'id' ] = KSIdGenerator.uuidInLowercaseHexWithHyphens(),
				[ 'type' ] = "array"
			}
		}
	}

	local requestOptions = {
		requestMethod = 'POST',
		returnHttpErrorsAsData = true,
		multipartRequest = true,
		requestCost = WFRxHttpClient.requestCostUploadNegative,
		requestPriority = WFRxHttpClient.requestPriorityHighest,
		requestHeaders = {
			{
				header = 'x-session-id',
				value = sessionId,
			}
		}
	}

	local fileName = KSPathUtils.leafName( filePath )

	local multipartRequestData = {
		{
			name = 'request',
			contentType = 'application/json',
			value = KSJsonEncoder.prettyPrint( request, { pretty = true } )
		},
		{
			name = 'gi_IMAGE',
			filePath = filePath,
			fileName = fileName
		}
	}

	local serverEndpoint = optArgs.prod and kFireflyServiceUrl_Prod or kFireflyServiceUrl_Stage

	local wfCatSpec = wfCatalog:ozCatalogSpec()

	return wfCatSpec:put( serverEndpoint, multipartRequestData, requestOptions )

end

--------------------------------------------------------------------------------

--[[!

	\brief Uses the Firefly outpainting service to expand the specified image.

	\param wfCatalog (WFCatalog)

	\param filePath (string) A file path to the image.

	\param size (table) A table that contains the new output width and height:
		{
			width = <number>
			height = <number
		}

	\param optArgs (?table) Optional arguments to modify the outpainting request:

		prod (?boolean): If true, will issue the request against prod. Defaults to stage.
		prompt (?string): Text-input to guide the firefly output.
		shift (?number): Shift position for outpainting. Must be from -1 to 1. Default is 0.
		dilate (?number): Dilate mask to have better boundary. Must be from 0.0 to 0.5 Default is 0.

	\return Observable which fires with the HTTP response from the service.
		See https://wiki.corp.adobe.com/display/DISysEng/MultiDiffusion+Service#MultiDiffusionService-Outpainting(withAspectRatio)
		for more information on the response.

--]]

function WFCatOrbit.requestExpand( wfCatalog, filePath, size, optArgs )

	check( "WFCatalog", "+string", "table", "?table" )
	
	optArgs = optArgs or {}

	return fireflyOutpaintingRequest( wfCatalog, filePath, size, optArgs )

end

--------------------------------------------------------------------------------
--@mark DII MetaCaf
--------------------------------------------------------------------------------

local function getHoleClassification( value )

	local classification = 'CAF'

	if value == 1 then
		classification = 'SuperCAF'
	elseif value == 2 then
		classification = 'CMGAN'
	elseif value == 3 then
		classification = 'Clio-MD'
	end

	return classification

end

--------------------------------------------------------------------------------

local outputContentTypes = {
	[ 'application/octet-stream' ] = true,
	[ 'image/png' ] = true,
	[ 'image/jpg' ] = true
}

--------------------------------------------------------------------------------

local function getDCFormat( extn )
	return string.format( "image/%s", extn )
end

--------------------------------------------------------------------------------

local function getLCFileExtension( path )

	local _, _, extn = string.find( path, ".+%.([^.]+)" )
	extn = string.lower( extn )

	return extn

end

--------------------------------------------------------------------------------

local function performMetaCAFRequest( wfCatalog, imagePath, maskPath, optArgs )

	-- See https://wiki.corp.adobe.com/display/DISysEng/MetaCAF+Service

	logger:infof( "performMetaCAFRequest inputImage: %s, inputMask, %s", imagePath, maskPath )

	if not imagePath or not WFFileUtils.exists( imagePath ) then

		return Observable.throw(
				string.format( "The specified input image does not exist: %s", imagePath ) )

	end

	if not maskPath or not WFFileUtils.exists( maskPath ) then

		return Observable.throw(
				string.format( "The specified input mask does not exist: %s", maskPath ) )

	end

	local requestHeaders = {
		{
			header = 'X-Api-Key',
			value = 'AdobeSenseiPredictServiceProdKey' -- FIXME eventually.
		},
		{
			header = 'X-analyzer-id',
			value = kDIIMetaCAFServiceId
		},
		{
			header = 'Prefer',
			value = 'respond-sync, wait=60'
		},
		{
			header = 'Cache-Control',
			value = 'no-cache'
		},
	}

	local requestOptions = {
		requestMethod = 'POST',
		returnHttpErrorsAsData = true,
		requestHeaders = requestHeaders,
		multipartRequest = true,
		requestCost = WFRxHttpClient.requestCostUploadNegative,
		requestPriority = WFRxHttpClient.requestPriorityHighest
	}

	local inputImageExtn = getLCFileExtension( imagePath )
	local inputMaskExtn = getLCFileExtension( maskPath )

	local inputFormat = getDCFormat( inputImageExtn )
	local maskFormat = getDCFormat( inputMaskExtn )
	local outputFormat = inputImageExtn == 'jxl' and "image/jxl" or "image/png"

	local contentAnalyzerRequests = {
		[ "sensei:name" ] = "MetaCAF",
		[ "sensei:invocation_mode" ] = "synchronous",
		[ "sensei:invocation_batch" ] = false,
		[ "sensei:in_response" ] = false,
		[ "sensei:engines" ] = {
			{
				[ "sensei:execution_info" ] = {
					[ "sensei:engine" ] = kDIIMetaCAFServiceName
				},
				[ "sensei:params" ] = {},
				[ "sensei:inputs" ] = {
					[ "input"] = {
						[ "dc:format" ] = inputFormat,
						[ "sensei:multipart_field_name" ] = "input_image"
					},
					[ "mask"] = {
						[ "dc:format" ] = maskFormat,
						[ "sensei:multipart_field_name" ] = "input_mask"
					}
				},
				[ "sensei:outputs" ] = {
					[ "spl:response" ] = {
						[ "dc:format" ] = "application/json",
						[ "sensei:multipart_field_name" ] = "metadata"
					},
					[ "output" ] = {
						[ "dc:format" ] = outputFormat,
						[ "sensei:multipart_field_name" ] = "output_image"
					},
				}
			}
		}
	}

	local multipartRequestData = {
		{
			name = 'contentAnalyzerRequests',
			contentType = 'application/json',
			value = KSJsonEncoder.prettyPrint( contentAnalyzerRequests, { pretty = true } )
		},
		{
			name = 'input_image',
			filePath = imagePath,
			fileName = KSPathUtils.leafName( imagePath )
		},
		{
			name = 'input_mask',
			filePath = maskPath,
			fileName = KSPathUtils.leafName( maskPath )
		}
	}

	local wfCatSpec = wfCatalog:ozCatalogSpec()

	local endpoint = optArgs.stage and kDIIMetaCAFUrl_Stage or kDIIMetaCAFUrl_Prod

	return wfCatSpec:put( endpoint, multipartRequestData, requestOptions )

		:map( function( parts, headers )

			local statusCode = KSTableUtils_derefOpt( headers, 'status_code' )

			if statusCode == '200' then

				for i, part in ipairs( parts ) do

					local partHeaders = KSTableUtils_derefOpt( part, 'headers' )

					local partData = KSTableUtils_derefOpt( part, 'data' )

					local holeClassificationInt =  KSTableUtils_derefOpt( partData, 'hole_classification' )

					if holeClassificationInt then

						local classification = getHoleClassification( holeClassificationInt )
						logger:infof( "Hole Classification: %s", classification )

					end

					local partHeadersContentType = KSTableUtils_derefOpt( partHeaders, 'Content-Type' )

					if outputContentTypes[ partHeadersContentType ] then

						return partData

					end

				end

			end

			return error( "The DII MetaCAF request could not be fulfilled. Status code: %s", statusCode )

		end )

end

--------------------------------------------------------------------------------

--[[!

	\brief Fills a hole using DII MetaCAF

	\param wfCatalog (WFCatalog)
	\param imagePath (string) A file path to the input image.
	\param maskPath (string) A file path to the mask.
	\param optArgs (optional table) Optional parameters to modify the request:
		{
			stage (boolean): If true, the request will run against stage.
		}

	\return Observable which fires with generated image data. Throws an error
		if the operation fails.

--]]

function WFCatOrbit.requestFillHoleWithMetaCAF_experimental( wfCatalog, imagePath, maskPath, optArgs )

	check( "WFCatalog", "+string", "+string", "?table" )

	optArgs = optArgs or {}

	return performMetaCAFRequest( wfCatalog, imagePath, maskPath, optArgs )

end

--------------------------------------------------------------------------------
--@mark DII SSM
--------------------------------------------------------------------------------

local function performServerSideMaskingRequest( wfCatalog, imagePath, optArgs )

	-- See https://wiki.corp.adobe.com/pages/viewpage.action?pageId=2947463333

	logger:infof( "performServerSideMaskingRequest inputImage: %s", imagePath )

	if not imagePath or not WFFileUtils.exists( imagePath ) then

		return Observable.throw(
				string.format( "The specified input image does not exist: %s", imagePath ) )

	end

	local imsClientId = WFImsAppIdentity.getClientId()

	local requestHeaders = {
		{
			header = 'Prefer',
			value = 'respond-sync, wait=60'
		},
		{
			header = 'Cache-Control',
			value = 'no-cache'
		},
		{
			header = 'X-Api-Key',
			value = imsClientId
		},
	}

	local requestOptions = {
		requestMethod = 'POST',
		returnHttpErrorsAsData = true,
		requestHeaders = requestHeaders,
		multipartRequest = true,
		requestCost = WFRxHttpClient.requestCostUploadNegative,
		requestPriority = WFRxHttpClient.requestPriorityHighest
	}

	local contentAnalyzerRequests = {
		[ "sensei:name" ] = "ACR Masking",
		[ "sensei:invocation_mode" ] = "synchronous",
		[ "sensei:invocation_batch" ] = false,
		[ "sensei:in_response" ] = false,
		[ "sensei:engines" ] = {
			{
				[ "sensei:execution_info" ] = {
					[ "sensei:engine" ] = kDIIServerSideMaskingServiceName
				},
				[ "sensei:params" ] = {
					 [ "request_types" ] = {
						"selectSubject",
						"selectSky",
						"selectSkin"
					}
				},
				[ "sensei:inputs" ] = {
					[ "image"] = {
						[ "dc:format" ] = "image/jpeg",
						[ "sensei:multipart_field_name" ] = "infile"
					}
				},
				[ "sensei:outputs" ] = {
					[ "masks" ] = {
						[ "dc:format" ] = "image/tiff",
						[ "sensei:multipart_field_name" ] = "outfiles"
					},
					[ "metadata" ] = {
						[ "dc:format" ] = "application/json",
						[ "sensei:multipart_field_name" ] = "metadata"
					}
				}
			}
		}
	}

	local multipartRequestData = {
		{
			name = 'contentAnalyzerRequests',
			contentType = 'application/json',
			value = KSJsonEncoder.prettyPrint( contentAnalyzerRequests, { pretty = true } )
		},
		{
			name = 'infile',
			filePath = imagePath,
			fileName = KSPathUtils.leafName( imagePath )
		}
	}

	-- Prep Output Folder

	local parentPath = KSPathUtils.parent( imagePath )
	local resultFolder = KSPathUtils.child( parentPath, "result" )
	WFFileUtils.mkdir( resultFolder )

	-- Issue the Request

	local wfCatSpec = wfCatalog:ozCatalogSpec()

	local endpoint = optArgs.stage and kDIIServerSideMaskingUrl_Stage or kDIIServerSideMaskingUrl_Prod

	return wfCatSpec:put( endpoint, multipartRequestData, requestOptions )

		:flatMap( function( parts, headers )

			local statusCode = KSTableUtils_derefOpt( headers, 'status_code' )

			local contentTypeStr = KSTableUtils_derefOpt( headers, 'Content-Type' )
			local contentType = contentTypeStr:match( "([^;]+)" ) -- Use only left-most content string.

			if statusCode == '200' and type( parts ) == 'table' then

				return Observable.from( parts )

			elseif contentType == 'application/json' then

				local response = KSJsonDecoder.decodeWebSafeJson( parts )

				statusCode = KSTableUtils_derefOpt( response, 'statuses', 1, 'invocations', 1, 'status' )
				local statusMessage = KSTableUtils_derefOpt( response, 'statuses', 1, 'invocations', 1, 'message' )

				return Observable.throw(
						string.format( "The masking request could not be fulfilled. Status code: %s, %s",
								statusCode, statusMessage ) )

			end

			return Observable.throw(
					string.format( "The masking request could not be fulfilled. Status code: %s",
							statusCode ) )

		end )

		:map( function( part )

			-- The id field of each returned object serves as a mapping to
			-- the filename of the corresponding mask file in the multipart
			-- response.

			local partHeaders = KSTableUtils_derefOpt( part, 'headers' )

			local partHeadersContentType = KSTableUtils_derefOpt( partHeaders, 'Content-Type' )

			if outputContentTypes[ partHeadersContentType ] then

				local partData = KSTableUtils_derefOpt( part, 'data' )

				local maskName = KSIdGenerator.uuidInLowercaseHex()

				local maskPath = KSPathUtils.child( resultFolder, maskName )

				local maskData = KSData.make( partData )
				KSData.saveToFile( maskData, maskPath )
				maskData:close()

				return maskPath

			end

			return nil

		end )

		:scan( function( paths, maskPath )

			if maskPath then
				paths[ #paths + 1 ] = maskPath
			end

			return paths

		end, {} )

		:takeLast( 1 )

end

--------------------------------------------------------------------------------

--[[!

	\brief Server Side Masking

	\param wfCatalog (WFCatalog)
	\param imagePath (string) A file path to the input image.
	\param optArgs (optional table) Optional parameters to modify the request:
		{
			stage (boolean): If true, the request will run against stage.
		}

	\return Observable which fires with the file paths of the generated masks.

--]]

function WFCatOrbit.requestServerSideMasking_experimental( wfCatalog, imagePath, optArgs )

	check( "WFCatalog", "+string", "?table" )

	optArgs = optArgs or {}

	return performServerSideMaskingRequest( wfCatalog, imagePath, optArgs )

end

--============================================================================--

return WFCatOrbit
