--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2023 Adobe
All Rights Reserved.

NOTICE: All information contained herein is, and remains
the property of Adobe and its suppliers, if any. The intellectual
and technical concepts contained herein are proprietary to Adobe
and its suppliers and are protected by all applicable intellectual
property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material
is strictly forbidden unless prior written permission is obtained
from Adobe.

------------------------------------------------------------------------------]]

	-- WF.core
local KSData = require 'KSData'
local KSIdGenerator = require 'KSIdGenerator'
local KSPathUtils = require 'KSPathUtils'
local KSTableUtils = require 'KSTableUtils'
local WFFileUtils = require 'WFFileUtils'
local WFPlatformInfo = require 'WFPlatformInfo'
local WFRxTestCase = require 'WFRxTestCase'

	-- WF.oz_client.managed_catalog
local WFCatTestSession = require 'WFCatTestSession'
local WFCatOrbit = require 'WFCatOrbit'

--============================================================================--

local WFCatOrbitTest = {}

--------------------------------------------------------------------------------

-- Grab your user Auth Token from https://senseiui.cloud.adobe.io/workflows in
-- order to run the autotag tests. We could avoid this, and enable the autotag
-- tests permanently if our IMS clients had the appropriate scopes.

local kAuthToken = [[TOKEN GOES HERE]]

--------------------------------------------------------------------------------

local testDataPath = KSPathUtils.child(
		KSPathUtils.getStandardFilePath( 'testdata' ), "import-samples" )

local testDataFolder1 = KSPathUtils.child( testDataPath, "folder-1" )

local fileName = "small.JPG"
local filePath = KSPathUtils.child( testDataFolder1, fileName )

local testMetaCAFDataPath = KSPathUtils.child(
		KSPathUtils.getStandardFilePath( 'testdata' ), "metacaf" )

--------------------------------------------------------------------------------

local function extractSenseiEngineAndFirstResult( parts, headers )

	if headers and headers.status_code and
			tonumber( headers.status_code ) == 200 then

		local tag = KSTableUtils.derefOpt(
				parts, 2, 'data', 'classifications', 1, 'classification_entries', 1, 'class_label' ) or {}

		return tag

	end

end

--------------------------------------------------------------------------------

function WFCatOrbitTest.DISABLED_rxTestRequestAutotagsForImage_testNightly( rxTest )

	-- To run this test, set kAuthToken and remove the "DISABLED_" prefix.

	rxTest:setTimeLimit( 10 )

	rxTest:expectSingleValue( WFCatOrbit.requestAutotagsForImage_experimental(
			kAuthToken, filePath, 10 )

		:map( extractSenseiEngineAndFirstResult ), 'sea' )

end

--------------------------------------------------------------------------------

function WFCatOrbitTest.DISABLED_rxTestRequestCategoryAutotagsForImage_testNightly( rxTest )

	-- To run this test, set kAuthToken and remove the "DISABLED_" prefix.

	rxTest:setTimeLimit( 10 )

	rxTest:expectSingleValue( WFCatOrbit.requestCategoryAutotagsForImage_experimental(
			kAuthToken, filePath, 10 )

		:map( extractSenseiEngineAndFirstResult ), 'Landscapes' )

end

--------------------------------------------------------------------------------

local function setUpTestCatalog( rxTest )

	local maybeIndexDefinitionsFile = WFPlatformInfo.isMobile() and 'LrmCatalogIndexDefinitions.lua' or nil
	local catalogTestSession = WFCatTestSession( rxTest, nil, maybeIndexDefinitionsFile )
	local catalog = rxTest:getFirstValue( catalogTestSession:catalog() )

	rxTest:finally( function()

		catalogTestSession:close()

	end )

	return catalog

end

--------------------------------------------------------------------------------

local function testRequestExpand( rxTest, manuallyVerify )

	rxTest:setTimeLimit( 30 )

	local wfCatalog = setUpTestCatalog( rxTest )

	local size = {
		height = 1024,
		width = 2048
	}

	local optArgs = {
		shift = -1,
		dilate = 0.1,
		prompt = "A futuristic city with a busy port. Photorealistic."
	}

	rxTest:expectSingleValue( WFCatOrbit.requestExpand( wfCatalog, filePath, size, optArgs )

		:map( function( parts, headers )

			local statusCode = KSTableUtils.derefOpt( headers, 'status_code' )
			local hasPartWithImageContent = false

			if statusCode == '200' then

				local testRunResultsPath

				if manuallyVerify then

					local resultsPath = KSPathUtils.child(
							KSPathUtils.getStandardFilePath( 'desktop' ), "expand" )

					testRunResultsPath = KSPathUtils.child(
							resultsPath, KSIdGenerator.uuidInLowercaseHex() )

					WFFileUtils.mkdir( testRunResultsPath )

				end

				for i, part in ipairs( parts ) do

					local partHeadersContentType = KSTableUtils.derefOpt( part, 'headers', 'Content-Type' )

					if partHeadersContentType == 'image/jpeg' then

						hasPartWithImageContent = true

						if manuallyVerify then

							local resultImagePath = KSPathUtils.child( testRunResultsPath, string.format( "expand_%d.jpg", i ) )

							local imgData = KSData.make( part.data )
							KSData.saveToFile( imgData, resultImagePath )
							imgData:close()

						end

					end

				end

			end

			return statusCode, hasPartWithImageContent

		end ), '200', true )

end

--------------------------------------------------------------------------------

-- FIXME [jandrew 2025-03-11]: Disabled due to instability.

function WFCatOrbitTest.DISABLED_rxTestRequestExpand_testNightly( rxTest )

	testRequestExpand( rxTest )

end

--------------------------------------------------------------------------------

function WFCatOrbitTest.DISABLED_rxTestRequestExpand_ManuallyVerifyResults_testNightly( rxTest )

	-- When enabled, this test case will write generated output images to the Desktop.
	-- This is useful for manually verifying the results, but not much else.

	local verifyManually = true

	testRequestExpand( rxTest, verifyManually )

end

--------------------------------------------------------------------------------

local function requestFillHoleWithMetaCAF( rxTest, inputImage, inputMask, manuallyVerify )

	rxTest:setTimeLimit( 30 )

	local wfCatalog = setUpTestCatalog( rxTest )

	local _, _, extn = string.find( inputImage, ".+%.([^.]+)" )
	extn = string.lower( extn )

	local optArgs = {
		stage = true
	}

	rxTest:expectSingleValue( WFCatOrbit.requestFillHoleWithMetaCAF_experimental(
			wfCatalog, inputImage, inputMask, optArgs )

		:map( function( fileData )

			if manuallyVerify then

				local resultsPath = KSPathUtils.child(
						KSPathUtils.getStandardFilePath( 'desktop' ), "metacaf" )

				local testRunResultsPath = KSPathUtils.child(
						resultsPath, KSIdGenerator.uuidInLowercaseHex() )

				WFFileUtils.mkdir( testRunResultsPath )

				local resultImagePath = KSPathUtils.child( testRunResultsPath,
						string.format( "expand_%s.%s", KSIdGenerator.uuidInLowercaseHex(), extn ) )

				local imgData = KSData.make( fileData )
				KSData.saveToFile( imgData, resultImagePath )
				imgData:close()

			end

			return not not fileData

		end ), true )

end

--------------------------------------------------------------------------------

function WFCatOrbitTest.rxTestRequestFillHoleWithMetaCAF_jxl_testNightly( rxTest )

	local fileName = "source.jxl"
	local inputImage = KSPathUtils.child( testMetaCAFDataPath, fileName )

	local maskFileName = "mask.jxl"
	local inputMask = KSPathUtils.child( testMetaCAFDataPath, maskFileName )

	requestFillHoleWithMetaCAF( rxTest, inputImage, inputMask )

end

--------------------------------------------------------------------------------

function WFCatOrbitTest.DISABLED_rxTestRequestFillHoleWithMetaCAF_ManuallyVerifyResults_testNightly( rxTest )

	local verifyManually = true

	local fileName = "source.jxl"
	local inputImage = KSPathUtils.child( testMetaCAFDataPath, fileName )

	local maskFileName = "mask.jxl"
	local inputMask = KSPathUtils.child( testMetaCAFDataPath, maskFileName )

	requestFillHoleWithMetaCAF( rxTest, inputImage, inputMask, verifyManually )

end

--------------------------------------------------------------------------------

-- TODO [jandrew 2024-07-26]: The following test has been disabled. The Orbit
-- entitlement service is blocking executing of the SSM request, due to our
-- test accounts not having proper authorization.

local function requestServerSideMasking( rxTest, inputImage )

	rxTest:setTimeLimit( 30 )

	local wfCatalog = setUpTestCatalog( rxTest )

	local optArgs = {
		stage = true
	}

	rxTest:expectSingleValue( WFCatOrbit.requestServerSideMasking_experimental(
			wfCatalog, inputImage, optArgs )

		:map( function( filePaths )

			return #filePaths > 0

		end ), true )

end

--------------------------------------------------------------------------------

function WFCatOrbitTest.rxTestRequestServerSideMasking_testNightly( rxTest )

	requestServerSideMasking( rxTest, filePath )

end

--------------------------------------------------------------------------------

return WFRxTestCase.wrapSuite( 'WFCatOrbitTest', WFCatOrbitTest )
