--[[----------------------------------------------------------------------------

WFMLEmbeddingsTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2025 Adobe
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe.

------------------------------------------------------------------------------]]

local WFFileUtils = require 'WFFileUtils'
local KSIdGenerator = require 'KSIdGenerator'
local WFPlatformInfo = require 'WFPlatformInfo'
local KSPathUtils = require "KSPathUtils"
local WFStruct = require "WFStruct"
local WFML = require "WFML"
local WFRxTestCase = require "WFRxTestCase"

local WFMLEmbeddingsTest = {}

if not WFPlatformInfo.isiOS then
	return WFMLEmbeddingsTest
end

function WFMLEmbeddingsTest.setUp()

	local documentDir = KSPathUtils.getStandardFilePath( 'documents' )
	local baseName = KSPathUtils.child( documentDir, 'embeddings' )

	WFFileUtils.delete( baseName .. '.idx' )
	WFFileUtils.delete( baseName .. '.dat' )

end

WFMLEmbeddingsTest.tearDown = WFMLEmbeddingsTest.setUp

--------------------------------------------------------------------------------

function WFMLEmbeddingsTest.rxTestEraseFind( rxTest )

	local documentDir = KSPathUtils.getStandardFilePath( 'documents' )
	local baseName = KSPathUtils.child( documentDir, 'embeddings' )

	local embeddings = WFML.WFMLEmbeddings( baseName, 2 )

	local id1 = KSIdGenerator.uuidInLowercaseHex()
	local v1 = WFStruct.pack( '2f', 1.0, 1.0 )

	embeddings:insert( id1, v1 )
	rxTest:expectEqual( true, embeddings:hasKey( id1 ) )

	local id2 = KSIdGenerator.uuidInLowercaseHex()
	local v2 = WFStruct.pack( '2f', 0.0, 0.0 )

	embeddings:insert( id2, v2 )
	rxTest:expectEqual( true, embeddings:hasKey( id2 ) )

	local id3 = KSIdGenerator.uuidInLowercaseHex()
	local v3 = WFStruct.pack( '2f', 1.0, 0.9 )

	embeddings:insert( id3, v3 )
	rxTest:expectEqual( true, embeddings:hasKey( id3 ) )

	local id4 = KSIdGenerator.uuidInLowercaseHex()
	local v4 = WFStruct.pack( '2f', 0.0, 1.0 )

	embeddings:insert( id4, v4 )
	rxTest:expectEqual( true, embeddings:hasKey( id4 ) )

	embeddings:erase( id1 )
	rxTest:expectEqual( false, embeddings:hasKey( id1 ) )

	local vs = WFStruct.pack( '2f', 0.9, 0.9 )

	local matches = embeddings:findClosest( vs, 2, 0.0 )

	rxTest:expectEqual( 2, #matches )
	rxTest:expectEqual( id3, matches[1].id )

end

--------------------------------------------------------------------------------

function WFMLEmbeddingsTest.rxTestInsertFind( rxTest )

	local documentDir = KSPathUtils.getStandardFilePath( 'documents' )
	local baseName = KSPathUtils.child( documentDir, 'embeddings' )

	local embeddings = WFML.WFMLEmbeddings( baseName, 2 )

	rxTest:expectEqual( 2, embeddings.vectorSize )
	rxTest:expectEqual( baseName, embeddings.baseName )

	local id1 = KSIdGenerator.uuidInLowercaseHex()
	local v1 = WFStruct.pack( '2f', 1.0, 1.0 )

	embeddings:insert( id1, v1 )
	rxTest:expectEqual( true, embeddings:hasKey( id1 ) )

	local id2 = KSIdGenerator.uuidInLowercaseHex()
	local v2 = WFStruct.pack( '2f', 0.0, 0.0 )

	embeddings:insert( id2, v2 )
	rxTest:expectEqual( true, embeddings:hasKey( id2 ) )

	local id3 = KSIdGenerator.uuidInLowercaseHex()
	local v3 = WFStruct.pack( '2f', 1.0, 0.0 )

	embeddings:insert( id3, v3 )
	rxTest:expectEqual( true, embeddings:hasKey( id3 ) )

	local id4 = KSIdGenerator.uuidInLowercaseHex()
	local v4 = WFStruct.pack( '2f', 0.0, 1.0 )

	embeddings:insert( id4, v4 )
	rxTest:expectEqual( true, embeddings:hasKey( id4 ) )

	local vs = WFStruct.pack( '2f', 0.9, 0.9 )

	local matches = embeddings:findClosest( vs, 2, 0.0 )

	rxTest:expectEqual( 2, #matches )
	rxTest:expectEqual( id1, matches[1].id )

end

--------------------------------------------------------------------------------

function WFMLEmbeddingsTest.rxTestInsertOffload( rxTest )

	local documentDir = KSPathUtils.getStandardFilePath( 'documents' )
	local baseName = KSPathUtils.child( documentDir, 'embeddings' )

	local embeddings = WFML.WFMLEmbeddings( baseName, 2 )

	rxTest:expectEqual( 2, embeddings.vectorSize )
	rxTest:expectEqual( baseName, embeddings.baseName )

	local id1 = KSIdGenerator.uuidInLowercaseHex()
	local v1 = WFStruct.pack( '2f', 1.0, 1.0 )

	embeddings:insert( id1, v1 )
	rxTest:expectEqual( true, embeddings:hasKey( id1 ) )

	local id2 = KSIdGenerator.uuidInLowercaseHex()
	local v2 = WFStruct.pack( '2f', 0.0, 0.0 )

	embeddings:insert( id2, v2 )
	rxTest:expectEqual( true, embeddings:hasKey( id2 ) )

	local id3 = KSIdGenerator.uuidInLowercaseHex()
	local v3 = WFStruct.pack( '2f', 1.0, 0.0 )

	embeddings:insert( id3, v3 )
	rxTest:expectEqual( true, embeddings:hasKey( id3 ) )

	embeddings:offloadIndex()

	local id4 = KSIdGenerator.uuidInLowercaseHex()
	local v4 = WFStruct.pack( '2f', 0.0, 1.0 )

	embeddings:insert( id4, v4 )
	rxTest:expectEqual( true, embeddings:hasKey( id4 ) )

	local vs = WFStruct.pack( '2f', 0.9, 0.9 )

	embeddings:offloadIndex()

	local matches = embeddings:findClosest( vs, 2, 0.0 )

	rxTest:expectEqual( 2, #matches )
	rxTest:expectEqual( id1, matches[1].id )

end

--------------------------------------------------------------------------------

return WFRxTestCase.wrapSuite( 'WFMLEmbeddingsTest', WFMLEmbeddingsTest )

