--[[----------------------------------------------------------------------------

WFTableArchivingTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2014 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local Rx = require 'Rx'

	-- WF.core.self_test
local WFRxTestCase = require 'WFRxTestCase'

	-- WF.core.table
local WFTableArchiving = require 'WFTableArchiving'

--============================================================================--

local WFTableArchivingTest = {}

--------------------------------------------------------------------------------

local function testTableToString( testName, t, expectedResult )

	WFTableArchivingTest[ "rxTestTableToString_" .. testName ] = function( rxTest )

		local result = WFTableArchiving.tableToString( t )

		assert( WFTableArchiving.isArchive( result ) )

		local KSTableArchiving = require 'KSTableArchiving'
		local result2 = KSTableArchiving.simpleTableToChunkString_deprecated( t )
		rxTest:expectEqual( result, result2 )

		rxTest:expectEqual( expectedResult, result )

	end

end

--------------------------------------------------------------------------------

local function testTableToStringError( testName, t )

	WFTableArchivingTest[ "rxTestTableToString_error_" .. testName ] = function( rxTest )

		rxTest:expectOnlyError( Rx.Observable.pcall( function()
			WFTableArchiving.tableToString( t )
		end ),
	':ignore:' )

	end
end

--------------------------------------------------------------------------------

testTableToString( 'arrayAfterFalse',
	{ 42, "hi", false, true, 5, 6, foo = { "foo" } },
	'{\n\t42,\n\t"hi",\n\tfalse,\n\ttrue,\n\t5,\n\t6,\n\tfoo = {\n\t\t"foo",\n\t},\n}' )

--------------------------------------------------------------------------------

testTableToString( 'arrayWithFloatKey',
	{ 42, 45, 98, 99, [ 3.14 ] = "oddball" },
	'{\n\t42,\n\t45,\n\t98,\n\t[3.14] = "oddball",\n\t99,\n}' )

--------------------------------------------------------------------------------

testTableToString( 'arrayWithNegativeKey',
	{ 42, 45, 98, 99, [ -87 ] = "oddball" },
	'{\n\t[-87] = "oddball",\n\t42,\n\t45,\n\t98,\n\t99,\n}' )

--------------------------------------------------------------------------------

testTableToString( 'arrayWithGap',
	{ 42, 45, 98, 99, nil, "six" },
	'{\n\t42,\n\t45,\n\t98,\n\t99,\n\t[6] = "six",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'emptyTable',
	{},
	'{\n}' )

--------------------------------------------------------------------------------

testTableToString( 'nestedEmptyTable',
	{ foo = {} },
	'{\n\tfoo = {\n\t},\n}' )

--------------------------------------------------------------------------------

testTableToString( 'keySorted_1_Table',
{ b = "beh", a = "ah" },
'{\n\ta = "ah",\n\tb = "beh",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'keySorted_2_Table',
{ a = "ah", b = "beh" },
'{\n\ta = "ah",\n\tb = "beh",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'keySorted_3_Table',
{ a1 = "a eins", a = "ah" },
'{\n\ta = "ah",\n\ta1 = "a eins",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'keySorted_4_Table',
{ a = "ah", a1 = "a eins" },
'{\n\ta = "ah",\n\ta1 = "a eins",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'emptyStringAsKey',
	{ [ "" ] = "empty", [ "haha" ] = "tricked you!" },
	'{\n\t[""] = "empty",\n\thaha = "tricked you!",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'nonIdentifierAsKey',
	{ [ "%" ] = "percent" },
	'{\n\t["%"] = "percent",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'leadingDigitInKey',
	{ [ "4sale" ] = "percent" },
	'{\n\t["4sale"] = "percent",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'internalDigitInKey',
	{ [ "sale4you" ] = "percent" },
	'{\n\tsale4you = "percent",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'keywordAsKey',
	{ [ "end" ] = "near" },
	'{\n\t["end"] = "near",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'embeddedZero',
	{ zero = "one\0zero" },
	'{\n\tzero = "one\\000zero",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'doesntUseQuoteSyntaxUnnecessarily',
	{ expires = "2014-03-13T21:46:17.705Z" },
	'{\n\texpires = "2014-03-13T21:46:17.705Z",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'stringContainingQuotes',
	{ zero = 'foo"bar"blah' },
	'{\n\tzero = "foo\\\"bar\\\"blah",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'nestedStringContainingQuotes',
	{
		error = {
			message = "[string \"WFSQLiteDatabase.lua\"]:915: WFDocStoreImp:setDocValue: _changeCounter :new: does not match existing document change counter 1",
			time = "2014-03-13T21:45:47.704Z",
		},
	},
	'{\n\terror = {\n\t\tmessage = "[string \\"WFSQLiteDatabase.lua\\"]:915: WFDocStoreImp:setDocValue: _changeCounter :new: does not match existing document change counter 1",\n\t\ttime = "2014-03-13T21:45:47.704Z",\n\t},\n}' )

--------------------------------------------------------------------------------

testTableToString( 'booleanTrueAsKey',
	{ [ true ] = 1 },
	'{\n\t[true] = 1,\n}' )

--------------------------------------------------------------------------------

testTableToString( 'booleanFalseAsKey',
	{ [ false ] = 0 },
	'{\n\t[false] = 0,\n}' )

--------------------------------------------------------------------------------

do
	local KSTableArchiving = require 'KSTableArchiving'
	local t = { }

	for i = 0, 255 do
		local s = require "WFZString".encodeUTF8Character( i )
		t[ s ] = s
	end
	local expected = KSTableArchiving.simpleTableToChunkString_deprecated( t )
	testTableToString( "char0thru255)", t, expected )
end

--------------------------------------------------------------------------------

testTableToString( 'embeddedQuotes',
	{ [ "\"" ] = "\"" },
	'{\n\t["\\\""] = "\\\"",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'variousNaNValues',
	{ 0 / 0, 1 / 0, -1 / 0, 42, "hi" },
	'{\n\t0 / 0,\n\t1 / 0,\n\t-1 / 0,\n\t42,\n\t"hi",\n}' )

--------------------------------------------------------------------------------

testTableToString( 'mightBlowLuaStack',
	{ {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}},
	  {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}},
	  {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}},
	  {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}},
	  {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, {{}}, },
	 '{\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n\t{\n\t\t{\n\t\t},\n\t},\n}'
	)

--------------------------------------------------------------------------------

testTableToStringError( 'invalidKeyType', { [ {} ] = "nope" } )

--------------------------------------------------------------------------------

testTableToStringError( 'invalidValueType', { function() end } )

--------------------------------------------------------------------------------

testTableToStringError( 'notTable', "42" )

--------------------------------------------------------------------------------

local KSUserdataProxy = assert( require 'KSUserdataProxy' )
local factory = KSUserdataProxy.makeFactory()
local proxy = factory()
assert( type( proxy ) == 'userdata' )

testTableToStringError( 'userdata', { a = 1, b = proxy, c = 2 } )

--------------------------------------------------------------------------------

--[[
function WFTableArchivingTest.OLD_testTableToString()

	-- Test with more values to increase coverage.

	simpleTable = { 0 / 0, 1 / 0, 42, "hi", false, true, foo = { "foo" } }
	simpleTable[ true ] = { 1 }
	simpleTable[ false ] = { 0 }

	resultString = WFTableArchiving.tableToString( simpleTable )

	local resultTable = loadstring( 'return ' .. resultString )()

	local resultString2 = WFTableArchiving.tableToString( resultTable )

	assert( resultString2 == resultString, "Table failed to reload and re-serialize the same" )

end
]]--

--------------------------------------------------------------------------------

local function _testTablePerformance_testPerformance( fn )

	for i = 1, 60 do

		local t = {}

		for o = 1, 100 do

			local t2 = {
				_id = string.format( "0000abcd0000abcd0000abcd%04d%04d", i, o ),
				_serverProvidedMetadata = {
					abcdefghijk = 12,
					[ "wicked \" key \"" ] = {
						twisted = true,
					},
					xmp = string.rep( "x", 1024 ),
				},
			}
			for u = 20, 128 do
				t2[ string.char( u ) ] = string.char( u )
			end

			t[ o ] = t2

		end

		fn( t )

	end

	collectgarbage( 'collect' )
	collectgarbage( 'collect' )
	collectgarbage( 'collect' )

end

--------------------------------------------------------------------------------

function WFTableArchivingTest.testNoOpPerformance_testPerformance( rxTest )

	_testTablePerformance_testPerformance( function() end )

end

--------------------------------------------------------------------------------

function WFTableArchivingTest.rxTestLuaTableArchivingPerformance_testPerformance( rxTest )
	local KSTableArchiving = require 'KSTableArchiving'
	_testTablePerformance_testPerformance( KSTableArchiving.simpleTableToChunkString_deprecated )

end

function WFTableArchivingTest.rxTestWFTableArchivingPerformance_testPerformance( rxTest )

	_testTablePerformance_testPerformance( WFTableArchiving.tableToString )

end

--------------------------------------------------------------------------------

return WFRxTestCase.wrapSuite( 'WFTableArchivingTest', WFTableArchivingTest )
